//Phong reflection model Fragment Shader
#version 430

in Data{
	vec3 pos; // in world space
	vec3 normal; // normalized and in world space
	vec2 texCoords;	
	float blur;
} dataIn;

layout (binding = 0) uniform  sampler2D tex;
layout (binding = 1) uniform samplerCube depthMap;
uniform bool shadows;
uniform float far_plane;
uniform int nearestLight; // index of the nearest light, that will cast shadows


// if no texture is used, set useTex = 0. Then the MaterialInfo below is used
// if a texture is present, set useTex = 1. Instead of the MaterialInfo, the texture is used for lightning
uniform int useTex;

// Material
layout(std140, binding = 0) uniform MaterialInfo{
// DO NOT CHANGE THIS code without changing buffer filling in Material.cpp
// using vec4 for conviniently using a c++ class to match the std140 layout. The fourth value in color only variables is ambigious
	vec4 diffuse;		
	vec4 specular;		
	vec4 ambient;		
	float shininess;
	float opacity;
} material;

// Lights
const int maxLights = 6;
struct LightInfo{
	vec4 pos;
	vec4 diffuse;
	vec4 specular;
	vec4 att; // constant, linear and quadratic attenuation, currently only affects point lights
};
layout(std140, binding = 1) uniform LightInfoBlock{	
	LightInfo lights[maxLights];
};
uniform int numLights;
uniform bool transparency;
uniform vec3 eyePos;

// the moving light
//uniform vec3 ml_pos;
//uniform vec3 ml_diff;
//uniform vec3 ml_spec;
//uniform vec3 ml_att;

// moving shadow
uniform float movingShadowPos; 
uniform float movingShadowRadius;

uniform bool gameover;

vec4 sceneAmbient = vec4(0.3, 0.3, 0.3, 1.0); // TODO replace with uniform

layout(location = 0) out vec4 color;

// info is used to store aditional information
// info.x stores the amount of blur per pixel
//
// the other components are not used currently
// if a component is used, name it here
layout(location = 1) out vec4 info;
// additional color attachments can be added in the framebuffer object

	// array of offset direction for sampling
vec3 gridSamplingDisk[20] = vec3[]
(
	vec3(1, 1, 1), vec3(1, -1, 1), vec3(-1, -1, 1), vec3(-1, 1, 1), 
	vec3(1, 1, -1), vec3(1, -1, -1), vec3(-1, -1, -1), vec3(-1, 1, -1),
	vec3(1, 1, 0), vec3(1, -1, 0), vec3(-1, -1, 0), vec3(-1, 1, 0),
	vec3(1, 0, 1), vec3(-1, 0, 1), vec3(1, 0, -1), vec3(-1, 0, -1),
	vec3(0, 1, 1), vec3(0, -1, 1), vec3(0, -1, -1), vec3(0, 1, -1)
);

float ShadowCalculation(vec3 fragPos, vec3 lightPos)
{
	

	vec3 fragToLight = fragPos - lightPos;
    // Get current linear depth as the length between the fragment and light position
    float currentDepth = length(fragToLight);
    // Test for shadows with PCF
    float shadow = 0.0;
    float bias = 0.15;
    //float bias = 0.3;
    int samples = 20;
    float viewDistance = length(eyePos - fragPos);
    float diskRadius = (1.0 + (viewDistance / far_plane)) / 25.0;
    for(int i = 0; i < samples; ++i)
    {
        float closestDepth = texture(depthMap, fragToLight + gridSamplingDisk[i] * diskRadius).r;
		//color = vec4(vec3(closestDepth / far_plane), 1.0); 
        closestDepth *= far_plane;   // Undo mapping [0;1]
        if(currentDepth - bias > closestDepth)
            shadow += 1.0;
    }
    shadow /= float(samples);
        
	return shadow;
}


void main(){	
	
	if(dataIn.pos.z <= movingShadowPos){
		// wrapped in total darkness
		color = vec4(0, 0, 0, 1);
		return;
	}
	
	float darknessFactor = min(1.0, (dataIn.pos.z - movingShadowPos)/movingShadowRadius);
	

	


	// define color parts either from the texture or the material info.
	int useMat = int(mod(useTex+1, 2)); // gives 0 if useTex is 1 and gives 1 if useTex is 0
	vec3 ambient = texture(tex, dataIn.texCoords).xyz * useTex + vec3(material.ambient) * useMat;
	vec3 diffuse = texture(tex, dataIn.texCoords).xyz * useTex + vec3(material.diffuse) * useMat;
	vec3 specular = vec3(1,1,1) * useTex + vec3(material.specular) * useMat; // using white for specular tex lightning TODO use something else
	float shininess = useTex + material.shininess * useMat;
	
	vec3 normal = normalize(dataIn.normal);
	vec3 viewDir = normalize(eyePos - dataIn.pos);
	vec3 lightDir;

	float attenuation;
	float dist;
	// TODO pack these into LightInfo
	float k0 = 0.1;
	float k1 = 0.0002;
	float k2 = 0.0006;
	
	vec3 akkColor = vec3(0, 0, 0);
	
	for(int i = 0; i < numLights; i++){
		float shadow = 0.0;
		if(lights[i].pos.w == 0){
			// directional light
			lightDir = normalize(vec3(lights[i].pos));
			attenuation = 1.0;

		} else {
			// point light
			
			// calculate shadow cast from nearest light
			// 1.. is in shadow
			// 0.. is not in shadow
			if(shadows && i == nearestLight){
				//shadow = shadows ? ShadowCalculation(dataIn.pos) : 0.0;  
				shadow = ShadowCalculation(dataIn.pos, lights[i].pos.xyz);
				
			}
			
			vec3 lightDirNotNormed = vec3(lights[i].pos) - dataIn.pos;
			dist = length(lightDirNotNormed);
			
			// if light is too far away, dont take it into account
			// TODO adjust distance depending on attenuation
			// TODO calculate distance in vertex shader
			if(dist > 50){
				continue;
			}

			lightDir = normalize(lightDirNotNormed);
			attenuation = 1.0;
			attenuation = 1.0 / (lights[i].att.x + (lights[i].att.y * dist) + (lights[i].att.z * dist * dist));			
		}
		
		// ambient lightning
		//vec3 ambientLightning = vec3(sceneAmbient) * ambient;
		
		// diffuse lightning
		vec3 diffuseReflection = vec3(lights[i].diffuse) * diffuse * max(0.0, dot(normal, lightDir));

		// specular lightning
		vec3 specularReflection = vec3(0.0, 0.0, 0.0); // no specular reflection if light source is on the wrong side
		// TODO try >= and look at it, discuss this!
		if(dot(normal, lightDir) > 0.0){			
			specularReflection = vec3(lights[i].specular) * specular
							* pow(max(0.0, dot(reflect(-lightDir, normal), viewDir)), shininess);
		}

		//akkColor = akkColor + (attenuation * (ambientLightning + (1.0 - shadow) * (diffuseReflection + specularReflection)));		
		akkColor = akkColor + (attenuation * ((1.0 - shadow) * (diffuseReflection + specularReflection)));		
	}

	// ambient lightning
	akkColor += vec3(sceneAmbient) * ambient;
	
	//////////////////////
	// the moving light //
	//////////////////////
	//vec3 lightDirNotNormed = vec3(ml_pos) - dataIn.pos;
	//lightDir = normalize(lightDirNotNormed);
	//dist = length(lightDirNotNormed);
	//attenuation = 1.0;
	//attenuation = 1.0 / (ml_att.x + (ml_att.y * dist) + (ml_att.z * dist * dist));

	//// ambient lightning
	//vec3 ambientLightning = vec3(sceneAmbient) * ambient;
		
	//// diffuse lightning
	//vec3 diffuseReflection = vec3(ml_diff) * diffuse * max(0.0, dot(normal, lightDir));

	//// specular lightning
	//vec3 specularReflection = vec3(0.0, 0.0, 0.0); // no specular reflection if light source is on the wrong side
	//// TODO try >= and look at it, discuss this!
	//if(dot(normal, lightDir) > 0.0){			
	//	specularReflection = vec3(ml_spec) * specular
	//					* pow(max(0.0, dot(reflect(-lightDir, normal), viewDir)), shininess);
	//}

	//akkColor = akkColor + (attenuation * (ambientLightning + diffuseReflection + specularReflection));		
	/////////////////////
	/////////////////////
	
	color = vec4(akkColor * darknessFactor, (transparency) ? material.opacity : 1.0) ;
	
	if(gameover)
		color += vec4(0.2, 0, 0, 0);

	// store amount of blur
	info = vec4(dataIn.blur, 1, 1, 1);
}